# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class GoogleSparkConnectException(Exception):
    """A custom exception class to only print the error messages.
    This would be used for exceptions where the stack trace
    doesn't provide any additional information.h
    """

    def __init__(self, error_message):
        self.error_message = error_message
        super().__init__()

    def _render_traceback_(self):
        print(self.error_message)



#!/bin/env python

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import contextlib
import logging
import socket
import threading
import time

import websockets.sync.client as websocketclient

from google import auth as googleauth
from google.auth.transport import requests as googleauthrequests

parser = argparse.ArgumentParser()
parser.add_argument("port")
parser.add_argument("target_host")


logger = logging.getLogger(__name__)


class bridged_socket(object):
    """Socket-like object that uses a websocket-over-TCP Bridge transport.

    See: https://github.com/google/inverting-proxy/tree/master/utils/tcpbridge
    """

    def __init__(self, websocket_conn):
        self._conn = websocket_conn

    def recv(self, buff_size):
        # N.B. The websockets [recv method](https://websockets.readthedocs.io/en/stable/reference/sync/client.html#websockets.sync.client.ClientConnection.recv)
        # does not support the buff_size parameter, but it does add a `timeout` keyword parameter not supported by normal
        # socket objects.
        #
        # We set that timeout to 60 seconds to prevent any scenarios where we wind up stuck waiting for a message from a websocket connection
        # that never comes.
        msg = self._conn.recv(timeout=60)
        return bytes.fromhex(msg)

    def send(self, msg_bytes):
        msg = bytes.hex(msg_bytes)
        self._conn.send(msg)

    def close(self):
        return self._conn.close()


def connect_tcp_bridge(hostname):
    """Create a socket-like connection to the given hostname using websocket.

    The backend server connected to over the websocket connection must be
    running the TCP-bridge backend corresponding to this frontend.

    Args:
        hostname: The hostname of the server running the TCP-bridge backend.

    Returns:
        A socket-like object with `recv` and `send` methods.
    """
    path = "tcp-over-websocket-bridge/35218cb7-1201-4940-89e8-48d8f03fed96"
    creds, _ = googleauth.default(
        scopes=["https://www.googleapis.com/auth/cloud-platform"]
    )
    creds.refresh(googleauthrequests.Request())

    return websocketclient.connect(
        f"wss://{hostname}/{path}",
        additional_headers={"Authorization": f"Bearer {creds.token}"},
        open_timeout=30,
    )


def forward_bytes(name, from_sock, to_sock):
    """Continuously stream bytes from the `from_sock` to the `to_sock`.

    This method terminates when either the `from_sock` is closed (causing
    it to return a Falsy value from its `recv` method), or the first time
    it hits an exception.

    This method is intended to be run in a separate thread of execution.

    Args:
        from_sock: A socket-like object to stream bytes from.
        to_sock: A socket-like object to stream bytes to.
    """
    while True:
        try:
            bs = from_sock.recv(1024)
            if not bs:
                to_sock.close()
                return
            while bs:
                try:
                    to_sock.send(bs)
                    bs = None
                except TimeoutError:
                    # On timeouts during a send, we retry just the send
                    # to make sure we don't lose any bytes.
                    pass
        except TimeoutError:
            # On timeouts during a receive, we retry the entire flow.
            pass
        except Exception as ex:
            logger.debug(f"[{name}] Exception forwarding bytes: {ex}")
            to_sock.close()
            return


def connect_sockets(conn_number, from_sock, to_sock):
    """Create a connection between the two given ports.

    This method continuously streams bytes in both directions between the
    given `from_sock` and `to_sock` socket-like objects.

    The caller is responsible for creating and closing the supplied socekts.
    """
    forward_name = f"{conn_number}-forward"
    t1 = threading.Thread(
        name=forward_name,
        target=forward_bytes,
        args=[forward_name, from_sock, to_sock],
        daemon=True,
    )
    t1.start()
    backward_name = f"{conn_number}-backward"
    t2 = threading.Thread(
        name=backward_name,
        target=forward_bytes,
        args=[backward_name, to_sock, from_sock],
        daemon=True,
    )
    t2.start()
    t1.join()
    t2.join()


def forward_connection(conn_number, conn, addr, target_host):
    """Create a connection to the target and forward `conn` to it.

    This method creates a socket-like object holding a connection to the given
    target host, and then continuously streams bytes in both directions between
    `conn` and that newly created connection.

    Both the supplied incoming connection (`conn`) and the created outgoing
    connection are automatically closed when this method terminates.

    This method should be run inside of a daemon thread so that it will not
    block program termination.
    """
    with conn:
        with connect_tcp_bridge(target_host) as websocket_conn:
            backend_socket = bridged_socket(websocket_conn)
            # Set a timeout on how long we will allow send/recv calls to block
            #
            # The code that reads and writes to this connection will retry
            # on timeouts, so this is a safe change.
            conn.settimeout(10)
            connect_sockets(conn_number, conn, backend_socket)


class DataprocSessionProxy(object):
    """A TCP proxy for forwarding requests to Dataproc Serverless Sessions.

    Spark Connect clients connect to this proxy using the h2c (without-SSL)
    protocol, and this proxy adds SSL by tunneling those connections over
    an HTTPS/WebSocket connection to the backend server.

    The tunneled requests are authenticated using the Google Application
    Default Credentials.
    """

    def __init__(self, port, target_host):
        self._port = port
        self._target_host = target_host
        self._started = False
        self._killed = False
        self._conn_number = 0

    @property
    def port(self):
        """The local port the proxy is listening on"""
        return self._port

    def start(self, daemon=True):
        """Start the proxy.

        By the time this method returns the proxy has already started listening
        on its local port will accept incoming connections.
        """
        if self._started:
            raise Exception("Dataproc session proxy already started")
        self._started = True
        s = threading.Semaphore(value=0)
        t = threading.Thread(target=self._run, args=[s], daemon=daemon)
        t.start()
        s.acquire()

    def _run(self, s):
        with socket.create_server(("127.0.0.1", self._port)) as frontend_socket:
            if self._port == 0:
                self._port = frontend_socket.getsockname()[1]
            s.release()
            while not self._killed:
                conn, addr = frontend_socket.accept()
                logger.debug(f"Accepted a connection from {addr}...")
                self._conn_number += 1
                threading.Thread(
                    target=forward_connection,
                    args=[self._conn_number, conn, addr, self._target_host],
                    daemon=True,
                ).start()

    def stop(self):
        """Stop the proxy."""
        self._killed = True


@contextlib.contextmanager
def dataproc_session_proxy(port, target_host):
    """Context manager for creating a Dataproc Session proxy.

    Usage:
        with dataproc_session_proxy(0, backend_hostname) as p:
           local_port = p.port
           ...

    Args:
        port: The local port to listen on. Use `0` to pick a free port.
        target_host: The backend to proxy connections to.

    Returns:
        A context manager wrapping a DataprocSessionProxy instance.
    """
    proxy = DataprocSessionProxy(port, target_host)
    try:
        proxy.start(daemon=False)
        yield proxy
    finally:
        proxy.stop()


if __name__ == "__main__":
    args = parser.parse_args()
    with dataproc_session_proxy(int(args.port), args.target_host) as p:
        print(f"Proxy listening on port {p.port}")
        try:
            while True:
                pass
        except KeyboardInterrupt:
            pass



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import atexit
import json
import logging
import os
import random
import string
import time
import datetime
from time import sleep
from typing import Any, cast, ClassVar, Dict, Optional

from google.api_core import retry
from google.api_core.future.polling import POLLING_PREDICATE
from google.api_core.client_options import ClientOptions
from google.api_core.exceptions import Aborted, FailedPrecondition, InvalidArgument, NotFound, PermissionDenied
from google.cloud.dataproc_v1.types import sessions

from google.cloud.spark_connect.client import DataprocChannelBuilder
from google.cloud.dataproc_v1 import (
    CreateSessionRequest,
    GetSessionRequest,
    Session,
    SessionControllerClient,
    SessionTemplate,
    TerminateSessionRequest,
)
from google.protobuf import text_format
from google.protobuf.text_format import ParseError
from pyspark.sql.connect.session import SparkSession
from pyspark.sql.utils import to_str

from google.cloud.spark_connect.exceptions import GoogleSparkConnectException

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class GoogleSparkSession(SparkSession):
    """The entry point to programming Spark with the Dataset and DataFrame API.

    A DataprocRemoteSparkSession can be used to create :class:`DataFrame`, register :class:`DataFrame` as
    tables, execute SQL over tables, cache tables, and read parquet files.

    Examples
    --------

    Create a Spark session with Google Spark Connect.

    >>> spark = (
    ...     GoogleSparkSession.builder
    ...         .appName("Word Count")
    ...         .googleSessionConfig(Session())
    ...         .getOrCreate()
    ... ) # doctest: +SKIP
    """

    _active_s8s_session_uuid: ClassVar[Optional[str]] = None
    _project_id = None
    _region = None
    _client_options = None
    _active_s8s_session_id: ClassVar[Optional[str]] = None

    class Builder(SparkSession.Builder):

        _dataproc_runtime_spark_version = {"3.0": "3.5.1", "2.2": "3.5.0"}

        _session_static_configs = [
            "spark.executor.cores",
            "spark.executor.memoryOverhead",
            "spark.executor.memory",
            "spark.driver.memory",
            "spark.driver.cores",
            "spark.eventLog.dir",
            "spark.history.fs.logDirectory",
        ]

        def __init__(self):
            self._options: Dict[str, Any] = {}
            self._channel_builder: Optional[DataprocChannelBuilder] = None
            self._dataproc_config: Optional[Session] = None
            self._project_id = os.environ.get("GOOGLE_CLOUD_PROJECT")
            self._region = os.environ.get("GOOGLE_CLOUD_REGION")
            self._client_options = ClientOptions(
                api_endpoint=os.environ.get(
                    "GOOGLE_CLOUD_DATAPROC_API_ENDPOINT",
                    f"{self._region}-dataproc.googleapis.com",
                )
            )

        def __apply_options(self, session: "SparkSession") -> None:
            with self._lock:
                self._options = {
                    key: value
                    for key, value in self._options.items()
                    if key not in self._session_static_configs
                }
                self._apply_options(session)

        def projectId(self, project_id):
            self._project_id = project_id
            return self

        def location(self, location):
            self._region = location
            self._client_options.api_endpoint = os.environ.get(
                "GOOGLE_CLOUD_DATAPROC_API_ENDPOINT",
                f"{self._region}-dataproc.googleapis.com",
            )
            return self

        def googleSessionConfig(self, dataproc_config: Session):
            with self._lock:
                self._dataproc_config = dataproc_config
                for k, v in dataproc_config.runtime_config.properties.items():
                    self._options[cast(str, k)] = to_str(v)
                return self

        def remote(self, url: Optional[str] = None) -> "SparkSession.Builder":
            if url:
                raise NotImplemented(
                    "GoogleSparkSession does not support connecting to an existing remote server"
                )
            else:
                return self

        def create(self) -> "SparkSession":
            raise NotImplemented(
                "GoogleSparkSession allows session creation only through getOrCreate"
            )

        def __create_spark_connect_session_from_s8s(
            self, session_response, session_name
        ) -> "SparkSession":
            GoogleSparkSession._active_s8s_session_uuid = session_response.uuid
            GoogleSparkSession._project_id = self._project_id
            GoogleSparkSession._region = self._region
            GoogleSparkSession._client_options = self._client_options
            spark_connect_url = session_response.runtime_info.endpoints.get(
                "Spark Connect Server"
            )
            spark_connect_url = spark_connect_url.replace("https", "sc")
            if not spark_connect_url.endswith("/"):
                spark_connect_url += "/"
            url = f"{spark_connect_url.replace('.com/', '.com:443/')};session_id={session_response.uuid};use_ssl=true"
            logger.debug(f"Spark Connect URL: {url}")
            self._channel_builder = DataprocChannelBuilder(
                url,
                is_active_callback=lambda: is_s8s_session_active(
                    session_name, self._client_options
                ),
            )

            assert self._channel_builder is not None
            session = GoogleSparkSession(connection=self._channel_builder)

            GoogleSparkSession._set_default_and_active_session(session)
            self.__apply_options(session)
            return session

        def __create(self) -> "SparkSession":
            with self._lock:

                if self._options.get("spark.remote", False):
                    raise NotImplemented(
                        "GoogleSparkSession does not support connecting to an existing remote server"
                    )

                from google.cloud.dataproc_v1 import SessionControllerClient

                dataproc_config: Session = self._get_dataproc_config()
                session_template: SessionTemplate = self._get_session_template()

                self._get_and_validate_version(
                    dataproc_config, session_template
                )

                spark_connect_session = self._get_spark_connect_session(
                    dataproc_config, session_template
                )

                if not spark_connect_session:
                    dataproc_config.spark_connect_session = {}
                os.environ["SPARK_CONNECT_MODE_ENABLED"] = "1"
                session_request = CreateSessionRequest()
                session_id = self.generate_dataproc_session_id()

                session_request.session_id = session_id
                dataproc_config.name = f"projects/{self._project_id}/locations/{self._region}/sessions/{session_id}"
                logger.debug(
                    f"Configurations used to create serverless session:\n {dataproc_config}"
                )
                session_request.session = dataproc_config
                session_request.parent = (
                    f"projects/{self._project_id}/locations/{self._region}"
                )

                logger.debug("Creating serverless session")
                GoogleSparkSession._active_s8s_session_id = session_id
                s8s_creation_start_time = time.time()
                try:
                    session_polling = retry.Retry(
                        predicate=POLLING_PREDICATE,
                        initial=5.0,  # seconds
                        maximum=5.0,  # seconds
                        multiplier=1.0,
                        timeout=600,  # seconds
                    )
                    print("Creating Spark session. It may take few minutes.")
                    if (
                        "GOOGLE_SPARK_CONNECT_SESSION_TERMINATE_AT_EXIT"
                        in os.environ
                        and os.getenv(
                            "GOOGLE_SPARK_CONNECT_SESSION_TERMINATE_AT_EXIT"
                        ).lower()
                        == "true"
                    ):
                        atexit.register(
                            lambda: terminate_s8s_session(
                                self._project_id,
                                self._region,
                                session_id,
                                self._client_options,
                            )
                        )
                    operation = SessionControllerClient(
                        client_options=self._client_options
                    ).create_session(session_request)
                    print(
                        f"Interactive Session Detail View:  https://console.cloud.google.com/dataproc/interactive/{self._region}/{session_id}?project={self._project_id}"
                    )
                    session_response: Session = operation.result(
                        polling=session_polling
                    )
                    if (
                        "DATAPROC_SPARK_CONNECT_ACTIVE_SESSION_FILE_PATH"
                        in os.environ
                    ):
                        file_path = os.environ[
                            "DATAPROC_SPARK_CONNECT_ACTIVE_SESSION_FILE_PATH"
                        ]
                        try:
                            session_data = {
                                "session_name": session_response.name,
                                "session_uuid": session_response.uuid,
                            }
                            os.makedirs(
                                os.path.dirname(file_path), exist_ok=True
                            )
                            with open(file_path, "w") as json_file:
                                json.dump(session_data, json_file, indent=4)
                        except Exception as e:
                            logger.error(
                                f"Exception while writing active session to file {file_path} , {e}"
                            )
                except (InvalidArgument, PermissionDenied) as e:
                    GoogleSparkSession._active_s8s_session_id = None
                    raise GoogleSparkConnectException(
                        f"Error while creating serverless session: {e.message}"
                    )
                except Exception as e:
                    GoogleSparkSession._active_s8s_session_id = None
                    raise RuntimeError(
                        f"Error while creating serverless session"
                    ) from e

                logger.debug(
                    f"Serverless session created: {session_id}, creation time taken: {int(time.time() - s8s_creation_start_time)} seconds"
                )
                return self.__create_spark_connect_session_from_s8s(
                    session_response, dataproc_config.name
                )

        def _get_exiting_active_session(self) -> Optional["SparkSession"]:
            s8s_session_id = GoogleSparkSession._active_s8s_session_id
            session_name = f"projects/{self._project_id}/locations/{self._region}/sessions/{s8s_session_id}"
            session_response = get_active_s8s_session_response(
                session_name, self._client_options
            )

            session = GoogleSparkSession.getActiveSession()
            if session is None:
                session = GoogleSparkSession._default_session

            if session_response is not None:
                print(
                    f"Using existing session: https://console.cloud.google.com/dataproc/interactive/{self._region}/{s8s_session_id}?project={self._project_id}, configuration changes may not be applied."
                )
                if session is None:
                    session = self.__create_spark_connect_session_from_s8s(
                        session_response, session_name
                    )
                return session
            else:
                logger.info(
                    f"Session: {s8s_session_id} not active, stopping previous spark session and creating new"
                )
                if session is not None:
                    session.stop()

                return None

        def getOrCreate(self) -> "SparkSession":
            with GoogleSparkSession._lock:
                session = self._get_exiting_active_session()
                if session is None:
                    session = self.__create()
                if session:
                    self.__apply_options(session)
                return session

        def _get_dataproc_config(self):
            dataproc_config = Session()
            if self._dataproc_config:
                dataproc_config = self._dataproc_config
                for k, v in self._options.items():
                    dataproc_config.runtime_config.properties[k] = v
            elif "DATAPROC_SPARK_CONNECT_SESSION_DEFAULT_CONFIG" in os.environ:
                filepath = os.environ[
                    "DATAPROC_SPARK_CONNECT_SESSION_DEFAULT_CONFIG"
                ]
                try:
                    with open(filepath, "r") as f:
                        dataproc_config = Session.wrap(
                            text_format.Parse(
                                f.read(), Session.pb(dataproc_config)
                            )
                        )
                except FileNotFoundError:
                    raise FileNotFoundError(f"File '{filepath}' not found")
                except ParseError as e:
                    raise ParseError(f"Error parsing file '{filepath}': {e}")
            if "COLAB_NOTEBOOK_RUNTIME_ID" in os.environ:
                dataproc_config.labels["colab-notebook-runtime-id"] = (
                    os.environ["COLAB_NOTEBOOK_RUNTIME_ID"]
                )
            if "COLAB_NOTEBOOK_KERNEL_ID" in os.environ:
                dataproc_config.labels["colab-notebook-kernel-id"] = os.environ[
                    "COLAB_NOTEBOOK_KERNEL_ID"
                ]
            return dataproc_config

        def _get_session_template(self):
            from google.cloud.dataproc_v1 import (
                GetSessionTemplateRequest,
                SessionTemplateControllerClient,
            )

            session_template = None
            if self._dataproc_config and self._dataproc_config.session_template:
                session_template = self._dataproc_config.session_template
                get_session_template_request = GetSessionTemplateRequest()
                get_session_template_request.name = session_template
                client = SessionTemplateControllerClient(
                    client_options=self._client_options
                )
                try:
                    session_template = client.get_session_template(
                        get_session_template_request
                    )
                except Exception as e:
                    logger.error(
                        f"Failed to get session template {session_template}: {e}"
                    )
                    raise
            return session_template

        def _get_and_validate_version(self, dataproc_config, session_template):
            trimmed_version = lambda v: ".".join(v.split(".")[:2])
            version = None
            if (
                dataproc_config
                and dataproc_config.runtime_config
                and dataproc_config.runtime_config.version
            ):
                version = dataproc_config.runtime_config.version
            elif (
                session_template
                and session_template.runtime_config
                and session_template.runtime_config.version
            ):
                version = session_template.runtime_config.version

            if not version:
                version = "3.0"
                dataproc_config.runtime_config.version = version
            elif (
                trimmed_version(version)
                not in self._dataproc_runtime_spark_version
            ):
                raise ValueError(
                    f"runtime_config.version {version} is not supported. "
                    f"Supported versions: {self._dataproc_runtime_spark_version.keys()}"
                )

            server_version = self._dataproc_runtime_spark_version[
                trimmed_version(version)
            ]
            import importlib.metadata

            google_connect_version = importlib.metadata.version(
                "google-spark-connect"
            )
            client_version = importlib.metadata.version("pyspark")
            version_message = f"Spark Connect: {google_connect_version} (PySpark: {client_version}) Session Runtime: {version} (Spark: {server_version})"
            logger.info(version_message)
            if trimmed_version(client_version) != trimmed_version(
                server_version
            ):
                logger.warning(
                    f"client and server on different versions: {version_message}"
                )
            return version

        def _get_spark_connect_session(self, dataproc_config, session_template):
            spark_connect_session = None
            if dataproc_config and dataproc_config.spark_connect_session:
                spark_connect_session = dataproc_config.spark_connect_session
            elif session_template and session_template.spark_connect_session:
                spark_connect_session = session_template.spark_connect_session
            return spark_connect_session

        def generate_dataproc_session_id(self):
            timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
            suffix_length = 6
            random_suffix = "".join(
                random.choices(
                    string.ascii_lowercase + string.digits, k=suffix_length
                )
            )
            return f"sc-{timestamp}-{random_suffix}"

    def _repr_html_(self) -> str:
        if not self._active_s8s_session_id:
            return """
            <div>No Active Dataproc Spark Session</div>
            """

        s8s_session = f"https://console.cloud.google.com/dataproc/interactive/{self._region}/{self._active_s8s_session_id}"
        ui = f"{s8s_session}/sparkApplications/applications"
        version = ""
        return f"""
        <div>
            <p><b>Spark Connect</b></p>

            <p><a href="{s8s_session}?project={self._project_id}">Serverless Session</a></p>
            <p><a href="{ui}?project={self._project_id}">Spark UI</a></p>
        </div>
        """

    def _remove_stoped_session_from_file(self):
        if "DATAPROC_SPARK_CONNECT_ACTIVE_SESSION_FILE_PATH" in os.environ:
            file_path = os.environ[
                "DATAPROC_SPARK_CONNECT_ACTIVE_SESSION_FILE_PATH"
            ]
            try:
                with open(file_path, "w"):
                    pass
            except Exception as e:
                logger.error(
                    f"Exception while removing active session in file {file_path} , {e}"
                )

    def stop(self) -> None:
        with GoogleSparkSession._lock:
            if GoogleSparkSession._active_s8s_session_id is not None:
                terminate_s8s_session(
                    GoogleSparkSession._project_id,
                    GoogleSparkSession._region,
                    GoogleSparkSession._active_s8s_session_id,
                    self._client_options,
                )

                self._remove_stoped_session_from_file()
                GoogleSparkSession._active_s8s_session_uuid = None
                GoogleSparkSession._active_s8s_session_id = None
                GoogleSparkSession._project_id = None
                GoogleSparkSession._region = None
                GoogleSparkSession._client_options = None

            self.client.close()
            if self is GoogleSparkSession._default_session:
                GoogleSparkSession._default_session = None
            if self is getattr(
                GoogleSparkSession._active_session, "session", None
            ):
                GoogleSparkSession._active_session.session = None


def terminate_s8s_session(
    project_id, region, active_s8s_session_id, client_options=None
):
    from google.cloud.dataproc_v1 import SessionControllerClient

    logger.debug(f"Terminating serverless session: {active_s8s_session_id}")
    terminate_session_request = TerminateSessionRequest()
    session_name = f"projects/{project_id}/locations/{region}/sessions/{active_s8s_session_id}"
    terminate_session_request.name = session_name
    state = None
    try:
        session_client = SessionControllerClient(client_options=client_options)
        session_client.terminate_session(terminate_session_request)
        get_session_request = GetSessionRequest()
        get_session_request.name = session_name
        state = Session.State.ACTIVE
        while (
            state != Session.State.TERMINATING
            and state != Session.State.TERMINATED
            and state != Session.State.FAILED
        ):
            session = session_client.get_session(get_session_request)
            state = session.state
            sleep(1)
    except NotFound:
        logger.debug(f"Session {active_s8s_session_id} already deleted")
    # Client will get 'Aborted' error if session creation is still in progress and
    # 'FailedPrecondition' if another termination is still in progress.
    # Both are retryable but we catch it and let TTL take care of cleanups.
    except (FailedPrecondition, Aborted):
        logger.debug(
            f"Session {active_s8s_session_id} already terminated manually or terminated automatically through session ttl limits"
        )
    if state is not None and state == Session.State.FAILED:
        raise RuntimeError("Serverless session termination failed")


def get_active_s8s_session_response(
    session_name, client_options
) -> Optional[sessions.Session]:
    get_session_request = GetSessionRequest()
    get_session_request.name = session_name
    try:
        get_session_response = SessionControllerClient(
            client_options=client_options
        ).get_session(get_session_request)
        state = get_session_response.state
    except Exception as e:
        logger.info(f"{session_name} deleted: {e}")
        return None
    if state is not None and (
        state == Session.State.ACTIVE or state == Session.State.CREATING
    ):
        return get_session_response
    return None


def is_s8s_session_active(session_name, client_options) -> bool:
    if get_active_s8s_session_response(session_name, client_options) is None:
        return False
    return True



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib.metadata
import warnings

from .session import GoogleSparkSession

old_package_name = "dataproc-spark-connect"
current_package_name = "google-spark-connect"
try:
    importlib.metadata.distribution(old_package_name)
    warnings.warn(
        f"Package '{old_package_name}' is already installed in your environment. "
        f"This might cause conflicts with '{current_package_name}'. "
        f"Consider uninstalling '{old_package_name}' and only install '{current_package_name}'."
    )
except:
    pass



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import find_namespace_packages, setup
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name="google-spark-connect",
    version="0.5.5",
    description="Google client library for Spark Connect",
    long_description=long_description,
    author="Google LLC",
    url="https://github.com/GoogleCloudDataproc/dataproc-spark-connect-python",
    license="Apache 2.0",
    packages=find_namespace_packages(include=["google.*"]),
    install_requires=[
        "google-api-core>=2.19.1",
        "google-cloud-dataproc>=5.18.0",
        "websockets",
        "pyspark[connect]>=3.5",
    ],
)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

import google
import grpc
from pyspark.sql.connect.client import ChannelBuilder

from . import proxy

logger = logging.getLogger(__name__)


class DataprocChannelBuilder(ChannelBuilder):
    """
    This is a helper class that is used to create a GRPC channel based on the given
    connection string per the documentation of Spark Connect.

    This implementation of ChannelBuilder uses `secure_authorized_channel` from the
    `google.auth.transport.grpc` package for authenticating secure channel.

    Examples
    --------
    >>> cb =  ChannelBuilder("sc://localhost")
    ... cb.endpoint

    >>> cb = ChannelBuilder("sc://localhost/;use_ssl=true;token=aaa")
    ... cb.secure
    True
    """

    def __init__(self, url, is_active_callback=None):
        self._is_active_callback = is_active_callback
        super().__init__(url)

    def toChannel(self) -> grpc.Channel:
        """
        Applies the parameters of the connection string and creates a new
        GRPC channel according to the configuration. Passes optional channel options to
        construct the channel.

        Returns
        -------
        GRPC Channel instance.
        """
        # TODO: Replace with a direct channel once all compatibility issues with
        # grpc have been resolved.
        return self._proxied_channel()

    def _proxied_channel(self) -> grpc.Channel:
        return ProxiedChannel(self.host, self._is_active_callback)

    def _direct_channel(self) -> grpc.Channel:
        destination = f"{self.host}:{self.port}"

        credentials, project = google.auth.default(
            scopes=["https://www.googleapis.com/auth/cloud-platform"]
        )
        # Get an HTTP request function to refresh credentials.
        request = google.auth.transport.requests.Request()
        # Create a channel.

        return google.auth.transport.grpc.secure_authorized_channel(
            credentials,
            request,
            destination,
            None,
            None,
            options=self._channel_options,
        )


class ProxiedChannel(grpc.Channel):

    def __init__(self, target_host, is_active_callback):
        self._is_active_callback = is_active_callback
        self._proxy = proxy.DataprocSessionProxy(0, target_host)
        self._proxy.start()
        self._proxied_connect_url = f"sc://localhost:{self._proxy.port}"
        self._wrapped = ChannelBuilder(self._proxied_connect_url).toChannel()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        ret = self._wrapped.__exit__(*args)
        self._proxy.stop()
        return ret

    def close(self):
        ret = self._wrapped.close()
        self._proxy.stop()
        return ret

    def _wrap_method(self, wrapped_method):
        if self._is_active_callback is None:
            return wrapped_method

        def checked_method(*margs, **mkwargs):
            if (
                self._is_active_callback is not None
                and not self._is_active_callback()
            ):
                logger.warning(f"Session is no longer active")
                raise RuntimeError(
                    "Session not active. Please create a new session"
                )
            return wrapped_method(*margs, **mkwargs)

        return checked_method

    def stream_stream(self, *args, **kwargs):
        return self._wrap_method(self._wrapped.stream_stream(*args, **kwargs))

    def stream_unary(self, *args, **kwargs):
        return self._wrap_method(self._wrapped.stream_unary(*args, **kwargs))

    def subscribe(self, *args, **kwargs):
        return self._wrap_method(self._wrapped.subscribe(*args, **kwargs))

    def unary_stream(self, *args, **kwargs):
        return self._wrap_method(self._wrapped.unary_stream(*args, **kwargs))

    def unary_unary(self, *args, **kwargs):
        return self._wrap_method(self._wrapped.unary_unary(*args, **kwargs))

    def unsubscribe(self, *args, **kwargs):
        return self._wrap_method(self._wrapped.unsubscribe(*args, **kwargs))
